# *** imports

# ** core
from typing import Dict, Any

# ** app
from .feature import FeatureContext
from .error import ErrorContext
from ..models.feature import Request
from ..commands import (
    Command,
    ModelObject,
    import_dependency,
    TiferetError
)
from ..commands.app import (
    AppInterface,
    AppRepository,
    ImportAppRepository
)
from ..commands.dependencies import create_injector, Injector


# *** contexts

# ** context: app_context
class AppContext(object):

    # * attribute: settings
    settings: Dict[str, Any]

    # * method: init
    def __init__(self, settings: Dict[str, Any] = {}):
        '''
        Initialize the application context.
        
        :param settings: The application settings.
        :type settings: dict
        '''
        
        # Assign settings attribute.
        self.settings = settings

    # * method: import_app_repo
    def import_app_repo(self) -> AppRepository:
        '''
        Import the app repository.

        :param module_path: The module path.
        :type module_path: str
        :param class_name: The class name.
        :type class_name: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The app repository.
        :rtype: AppRepository
        '''

        # Run and return the import app repository command.
        return Command.handle(
            ImportAppRepository,
            **self.settings
        )
    
    # * method: get_interface_settings
    def get_interface_settings(self, interface_id: str) -> AppInterface:
        '''
        Get the settings for the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :return: The application interface settings.
        :rtype: AppInterface
        '''

        # Import the app repository.
        app_repo = self.import_app_repo()

        # Get the app interface.
        return app_repo.get_interface(interface_id)

    # ** method: create_app_injector
    def create_app_injector(self, app_interface: AppInterface) -> Injector:
        '''
        Create the app dependency injector.

        :param app_interface: The app interface.
        :type app_interface: AppInterface
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The dependencies injector.
        :rtype: Injector
        '''

        # Retrieve the app context dependency.
        dependencies = dict(
            app_context=import_dependency.execute(
                app_interface.module_path,
                app_interface.class_name,
            )
        )

        # Add the remaining app context attributes.
        for attr in app_interface.attributes:
            dependencies[attr.attribute_id] = import_dependency.execute(
                attr.module_path,
                attr.class_name,
            )

        # Create the injector.
        injector = create_injector.execute(
            app_interface.id, 
            dependencies,
            interface_id=app_interface.id,
            **app_interface.constants
        )

        # Return the injector.
        return injector
    
       # * method: load_interface
    def load_interface(self, interface_id: str) -> 'AppInterfaceContext':
        '''
        Load the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :return: The application interface context.
        :rtype: AppInterfaceContext
        '''

        # Get the app interface.
        app_interface = self.get_interface_settings(interface_id)

        # Create the app injector.
        injector = self.create_app_injector(app_interface)

        # Load the app interface context.
        return getattr(injector, 'app_context')

    # * method: run
    def run(self,
            interface_id: str,
            feature_id: str,
            headers: Dict[str, str] = {},
            data: Dict[str, Any] = {},
            debug: bool = False,
            **kwargs
        ) -> Any:
        '''
        Run the application interface.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param dependencies: The dependencies.
        :type dependencies: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The response.
        :rtype: Any
        '''

        # Load the interface.
        app_interface = self.load_interface(interface_id)

        # Run the interface.
        return app_interface.run(
            feature_id, 
            headers, 
            data, 
            debug=debug,
            **kwargs
        )


# ** context: app_interface_context
class AppInterfaceContext(object): 
    '''
    The application interface context is a class that is used to create and run the application interface.
    '''

    # * attribute: interface_id
    interface_id: str

    # * attribute: features
    features: FeatureContext

    # * attribute: errors
    errors: ErrorContext

    # * method: init
    def __init__(self, interface_id: str, features: FeatureContext, errors: ErrorContext):
        '''
        Initialize the application interface context.

        :param interface_id: The interface ID.
        :type interface_id: str
        :param app_name: The application name.
        :type app_name: str
        :param features: The feature context.
        :type features: FeatureContext
        :param errors: The error context.
        :type errors: ErrorContext
        '''

        # Assign instance variables.
        self.interface_id = interface_id
        self.features = features
        self.errors = errors

    # * method: parse_request
    def parse_request(self, headers: Dict[str, str] = {}, data: Dict[str, Any] = {}) -> Request:
        '''
        Parse the incoming request.

        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :return: The parsed request.
        :rtype: Request
        '''

        # Add the interface id to the request headers.
        headers.update(dict(
            interface_id=self.interface_id,
        ))

        # Create the request model object.
        request = ModelObject.new(
            Request,
            headers=headers,
            data=data,
        )

        # Return the request model object.
        return request
    
    # * method: execute_feature
    def execute_feature(self, feature_id: str, request: Request, **kwargs):
        '''
        Execute the feature context.

        :param feature_id: The feature identifier.
        :type feature_id: str
        :param request: The request.
        :type request: Request
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Add the feature id to the request headers.
        request.headers.update(dict(
            feature_id=feature_id
        ))

        # Execute feature context and return session.
        self.features.execute_feature(feature_id, request, **kwargs)

    # * method: handle_error
    def handle_error(self, error: Exception) -> Any:
        '''
        Handle the error and return the response.

        :param error: The error to handle.
        :type error: Exception
        :return: The error response.
        :rtype: Any
        '''
        
        # Print the error to the console.
        print('Error:', error)

        # Handle the error and return the response.
        return self.errors.handle_error(error)

    # * method: handle_response
    def handle_response(self, request: Request) -> Any:
        '''
        Handle the response from the request.

        :param request: The request context.
        :type request: RequestContext
        :return: The response.
        :rtype: Any
        '''

        # Handle the response and return it.
        return request.handle_response()
    
    # * method: run
    def run(self, 
            feature_id: str, 
            headers: Dict[str, str] = {}, 
            data: Dict[str, Any] = {},
            **kwargs) -> Any:
        '''
        Run the application interface by executing the feature.

        :param feature_id: The feature identifier.
        :type feature_id: str
        :param headers: The request headers.
        :type headers: dict
        :param data: The request data.
        :type data: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''
        
        # Parse request.
        request = self.parse_request(headers, data)

        # Execute feature context and return session.
        try:
            self.execute_feature(feature_id, request, **kwargs)

        # Handle error and return response if triggered.
        except TiferetError as e:
            return self.handle_error(e)

        # Handle response.
        return self.handle_response(request)



# *** imports

# ** core
from typing import Any
import json

# ** app
from .settings import *


# *** models

# ** model: request
class Request(ValueObject):
    '''
    A request object.
    '''

    # * attribute: headers
    headers = DictType(
        StringType(),
        metadata=dict(
            description='The request headers.'
        )
    )

    # * attribute: data
    data = DictType(
        StringType(),
        metadata=dict(
            description='The request data.'
        )
    )

    # * attribute: result
    result = StringType(
        metadata=dict(
            description='The request result.'
        )
    )

    # * method: set_result
    def set_result(self, result: Any):
        # Set the result as a serialized empty dictionary if it is None.
        if not result:
            self.result = json.dumps({})
            return
            
        # If the result is a Model, convert it to a primitive dictionary and serialize it.
        if isinstance(result, ModelObject):
            self.result = json.dumps(result.to_primitive())
            return

        # If the result is not a list, it must be a dict, so serialize it and set it.
        if type(result) != list:
            self.result = json.dumps(result)
            return

        # If the result is a list, convert each item to a primitive dictionary.
        result_list = []
        for item in result:
            if isinstance(item, ModelObject):
                result_list.append(item.to_primitive())
            else:
                result_list.append(item)

        # Serialize the result and set it.
        self.result = json.dumps(result_list)

    # * method: handle_response
    def handle_response(self, **kwargs) -> Any:
        '''
        Handle the response.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The response object.
        :rtype: Any
        '''

        # Deserialize the result.
        # Return None if the result is None.
        return json.loads(self.result) if self.result else None


# ** model: feature_command
class FeatureCommand(ValueObject):
    '''
    A command object for a feature command.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the feature handler.'
        )
    )

    # * attribute: attribute_id
    attribute_id = StringType(
        required=True,
        metadata=dict(
            description='The container attribute ID for the feature command.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType(),
        default={},
        metadata=dict(
            description='The custom parameters for the feature handler.'
        )
    )

    # * attribute: return_to_data (obsolete)
    return_to_data = BooleanType(
        default=False,
        metadata=dict(
            description='Whether to return the feature command result to the feature data context.'
        )
    )

    # * attribute: data_key
    data_key = StringType(
        metadata=dict(
            description='The data key to store the feature command result in if Return to Data is True.'
        )
    )

    # * attribute: pass_on_error
    pass_on_error = BooleanType(
        metadata=dict(
            description='Whether to pass on the error if the feature handler fails.'
        )
    )


# ** model: feature
class Feature(Entity):
    '''
    A feature object.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the feature.'
        )
    )

    # * attribute: group_id
    group_id = StringType(
        required=True,
        metadata=dict(
            description='The context group identifier for the feature.'
        )
    )

    feature_key = StringType(
        required=True,
        metadata=dict(
            description='The key of the feature.'
        )
    )

    # * attribute: commands
    commands = ListType(
        ModelType(FeatureCommand),
        default=[],
        metadata=dict(
            description='The command handler workflow for the feature.'
        )
    )

    # * attribute: log_params
    log_params = DictType(
        StringType(),
        default={},
        metadata=dict(
            description='The parameters to log for the feature.'
        )
    )

    # * method: new
    @staticmethod
    def new(name: str, group_id: str, feature_key: str = None, id: str = None, description: str = None, **kwargs) -> 'Feature':
        '''Initializes a new Feature object.

        :param name: The name of the feature.
        :type name: str
        :param group_id: The context group identifier of the feature.
        :type group_id: str
        :param feature_key: The key of the feature.
        :type feature_key: str
        :param id: The identifier of the feature.
        :type id: str
        :param description: The description of the feature.
        :type description: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new Feature object.
        '''

        # Set the feature key as the snake case of the name if not provided.
        if not feature_key:
            feature_key = name.lower().replace(' ', '_')

        # Feature ID is the group ID and feature key separated by a period.
        if not id:
            id = f'{group_id}.{feature_key}'

        # Set the description as the name if not provided.
        if not description:
            description = name

        # Create and return a new Feature object.
        return Entity.new(
            Feature,
            id=id,
            name=name,
            group_id=group_id,
            feature_key=feature_key,
            description=description,
            **kwargs
        )
    
    # * method: add_command
    def add_command(self, command: FeatureCommand, position: int = None):
        '''Adds a service command to the feature.

        :param command: The service command to add.
        :type command: FeatureCommand
        :param position: The position to add the handler at.
        :type position: int
        '''

        # Add the feature command to the feature.
        if position is not None:
            self.commands.insert(position, command)
        else:
            self.commands.append(command)



# *** imports

# ** infra
from schematics.types.serializable import serializable

# 
from ..data import DataObject
from ..contracts.feature import Feature as FeatureContract, FeatureCommand as FeatureCommandContract
from ..models.feature import *


class FeatureCommandData(FeatureCommand, DataObject):
    '''
    A data representation of a feature handler.
    '''

    class Options():
        '''
        The default options for the feature handler data.
        '''

        # Set the serialize when none flag to false.
        serialize_when_none = False

        # Define the roles for the feature handler data.
        roles = {
            'to_model': DataObject.deny('parameters'),
            'to_data': DataObject.allow()
        }

    # * attributes
    parameters = DictType(
        StringType(),
        default={},
        serialized_name='params',
        deserialize_from=['params', 'parameters'],
        metadata=dict(
            description='The parameters for the feature.'
        )
    )

    def map(self, role: str = 'to_model', **kwargs) -> FeatureCommandContract:
        '''
        Maps the feature handler data to a feature handler object.
        
        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature handler object.
        :rtype: f.FeatureCommand
        '''
        return super().map(FeatureCommand, 
            role, 
            parameters=self.parameters,
            **kwargs)


class FeatureData(Feature, DataObject):
    '''
    A data representation of a feature.
    '''

    class Options():
        '''
        The default options for the feature data.
        '''

        # Set the serialize when none flag to false.
        serialize_when_none = False

        # Define the roles for the feature data.
        roles = {
            'to_model': DataObject.deny('feature_key'),
            'to_data': DataObject.deny('feature_key', 'group_id', 'id')
        }
    
    # * attributes
    commands = ListType(ModelType(FeatureCommandData),
                          deserialize_from=['handlers', 'functions', 'commands'],)
    
    @serializable
    def feature_key(self):
        '''
        Gets the feature key.
        '''

        # Return the feature key.
        return self.id.split('.')[-1]

    def map(self, role: str = 'to_model', **kwargs) -> FeatureContract:
        '''
        Maps the feature data to a feature object.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new feature object.
        :rtype: f.Feature
        '''

        # Map the feature data to a feature object.
        return super().map(Feature, role, 
            feature_key=self.feature_key,
            commands=[
                command.map(role, **kwargs) for command in self.commands
            ],
            **kwargs
        )

    @staticmethod
    def from_data(**kwargs) -> 'FeatureData':
        '''
        Initializes a new FeatureData object from a Feature object.
        
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new FeatureData object.
        :rtype: FeatureData
        '''

        # Create a new FeatureData object.
        return super(FeatureData, FeatureData).from_data(
            FeatureData, 
            **kwargs
        )



# *** imports

# ** core
from typing import Any, Dict

# ** infra
from dependencies import Injector
from dependencies.exceptions import DependencyError

# ** app
from ..commands import *


# *** commands

# ** command: create_injector
class CreateInjector(Command):
    '''
    A command to create a dependencies (library) injector object with the given dependencies.
    '''

    def execute(self, name: str, dependencies: Dict[str, type], **kwargs) -> Any:
        '''
        Execute the command to create an injector object.

        :param name: The name of the injector.
        :type name: str
        :param dependencies: The dependencies.
        :type dependencies: dict
        :return: The injector object.
        :rtype: Any
        '''
    
        # Create container.
        return type(name, (Injector,), {**dependencies, **kwargs})


# ** command: get_dependency
class GetDependency(Command):
    '''
    A command to get a dependency from the injector.
    '''

    def execute(self, injector: Injector, dependency_name: str) -> Any:
        '''
        Execute the command to get a dependency from the injector.

        :param injector: The injector object.
        :type injector: Injector
        :param dependency_name: The name of the dependency to get.
        :type dependency_name: str
        :return: The dependency object.
        :rtype: Any
        '''
    
        # Return the dependency from the injector.
        try:
            return getattr(injector, dependency_name)
        
        # If the dependency does not exist or cannot be resolved, raise an error.
        except DependencyError as e:
            self.raise_error(
                'INVALID_DEPENDENCY_ERROR',
                f'Dependency {dependency_name} could not be resolved: {str(e)}',
                dependency_name,
                str(e)
            )


# *** command_variables

# ** command_variable: create_injector
create_injector = CreateInjector()

# ** command_variable: get_dependency
get_dependency = GetDependency()


# *** imports

# ** app
from ..data import *
from ..models.container import *
from ..contracts.container import (
    ContainerAttribute as ContainerAttributeContract, 
    FlaggedDependency as FlaggedDependencyContract
)


# *** data

# ** data: flagged_dependency_yaml_data
class FlaggedDependencyYamlData(FlaggedDependency, DataObject):
    '''
    A data representation of a flagged dependency object.
    '''

    class Options():
        '''
        The options for the flagged dependency data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('params'),
            'to_data': DataObject.deny('flag')
        }

    # * attribute: flag
    flag = StringType(
        metadata=dict(
            description='The flag is no longer required due to the YAML format.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType, 
        default={}, 
        serialized_name='params', 
        deserialize_from=['params'],
        metadata=dict(
            description='The parameters need to now account for new data names in the YAML format.'
        ),
    )

    # * method: map
    def map(self, **kwargs) -> FlaggedDependencyContract:
        '''
        Maps the flagged dependency data to a flagged dependency object.

        :param role: The role for the mapping.
        :type role: str
        :return: A new flagged dependency object.
        :rtype: FlaggedDependencyContract
        '''

        # Map to the container dependency object.
        obj = super().map(FlaggedDependency, **kwargs, validate=False)
    
        # Set the parameters in due to the deserializer.
        obj.parameters = self.parameters

        # Validate and return the object.
        obj.validate()
        return obj
    
    # * method: new
    @staticmethod
    def from_data(**kwargs) -> 'FlaggedDependencyYamlData':
        '''
        Initializes a new ContainerDependencyYamlData object from YAML data.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ContainerDependencyYamlData object.
        :rtype: ContainerDependencyYamlData
        '''

        # Create a new ContainerDependencyYamlData object.
        return super(
            FlaggedDependencyYamlData, 
            FlaggedDependencyYamlData
        ).from_data(
            FlaggedDependencyYamlData,
            **kwargs
        )
    
    # * method: from_model
    @staticmethod
    def from_model(model: FlaggedDependency, **kwargs) -> 'FlaggedDependencyYamlData':
        '''
        Initializes a new ContainerDependencyYamlData object from a model object.

        :param model: The flagged dependency model object.
        :type model: FlaggedDependency
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Create and return a new FlaggedDependencyYamlData object.
        return super(FlaggedDependencyYamlData, FlaggedDependencyYamlData).from_model(
            FlaggedDependencyYamlData,
            model,
            **kwargs,
        )


# ** data: container_attribute_yaml_data
class ContainerAttributeYamlData(ContainerAttribute, DataObject):
    '''
    A data representation of a container attribute object.
    '''

    class Options():
        '''
        The options for the container attribute data.
        '''

        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('params'),
            'to_data': DataObject.deny('id')
        }

    # * attribute: dependencies
    dependencies = DictType(
        ModelType(FlaggedDependencyYamlData), 
        default={}, 
        serialized_name='deps', 
        deserialize_from=['deps', 'dependencies'],
        metadata=dict(
            description='The dependencies are now a key-value pair keyed by the flags.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType, 
        default={}, 
        serialized_name='params', 
        deserialize_from=['params'],
        metadata=dict(
            description='The default parameters for the container attribute.'
        ),
    )

    # * method: map
    def map(self, **kwargs) -> ContainerAttributeContract:
        '''
        Maps the container attribute data to a container attribute object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A container attribute model contract.
        :rtype: ContainerAttributeContract
        '''

        # Map to the container attribute object with the dependencies.
        return super().map(ContainerAttribute, 
            dependencies=[dep.map(flag=flag) for flag, dep in self.dependencies.items()],
            parameters=self.parameters,
            **kwargs)

    # * method: new
    @staticmethod
    def from_data(**kwargs) -> 'ContainerAttributeYamlData':
        '''
        Initializes a new ContainerAttributeYamlData object from YAML data.

        :param deps: The dependencies data.
        :type deps: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''        

        # Create a new ContainerAttributeYamlData object.
        obj = super(
            ContainerAttributeYamlData, 
            ContainerAttributeYamlData
        ).from_data(
            ContainerAttributeYamlData,
            **kwargs, 
            validate=False
        )
    
        # Set the dependencies.
        for flag, dep in obj.dependencies.items():
            dep.flag = flag

        # Validate and return the object.
        obj.validate()
        return obj
    
    # * method: from_model
    @staticmethod
    def from_model(model: ContainerAttribute, **kwargs) -> 'ContainerAttributeYamlData':
        '''
        Initializes a new ContainerAttributeYamlData object from a model object.

        :param model: The container attribute model object.
        :type model: ContainerAttribute
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Create the dependency data.
        dependencies = {dep.flag: dep.to_primitive() for dep in model.dependencies}
        
        # Create a new model object without the dependencies.
        data = model.to_primitive()
        data['dependencies'] = dependencies

        # Create a new ContainerAttributeYamlData object.
        obj = ContainerAttributeYamlData(
            dict(
                **data,
                **kwargs
            ), 
            strict=False
        )

        # Validate and return the object.
        obj.validate()
        return obj


# *** imports

# ** app
from ..configs import *
from ..data import DataObject
from ..models.error import *
from ..contracts.error import Error as ErrorContract


# *** data

# ** data: error_data
class ErrorData(Error, DataObject):
    '''
    A data representation of an error object.
    '''

    class Options():
        serialize_when_none = False
        roles = {
            'to_data': DataObject.deny('id', 'message'),
            'to_model': DataObject.deny('message')
        }

    # * attribute: message
    message = ListType(
        ModelType(ErrorMessage),
        required=True,
        metadata=dict(
            description='The error messages.'
        )
    )

    # * to_primitive
    def to_primitive(self, role: str = 'to_data', **kwargs) -> dict:
        '''
        Converts the data object to a primitive dictionary.

        :param role: The role.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The primitive dictionary.
        :rtype: dict
        '''

        # Convert the data object to a primitive dictionary.
        return dict(
            **super().to_primitive(
                role,
                **kwargs
            ),
            message=[msg.to_primitive() for msg in self.message]
        ) 
        

    # * method: map
    def map(self, **kwargs) -> ErrorContract:
        '''
        Maps the error data to an error object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new error object.
        :rtype: Error
        '''

        # Map the error messages.
        return super().map(Error,
            **self.to_primitive('to_model'),
            **kwargs)

    # * method: from_data
    @staticmethod
    def from_data(**kwargs) -> 'ErrorData':
        '''
        Creates a new ErrorData object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ErrorData object.
        :rtype: ErrorData
        '''

        # Create a new ErrorData object.
        return super(ErrorData, ErrorData).from_data(
            ErrorData, 
            **kwargs
        )



# *** imports

# ** core
from typing import List, Dict, Any

# ** app
from .settings import *


# *** contracts

# ** contract: app_attribute
class AppAttribute(ModelContract):
    '''
    An app dependency contract that defines the dependency attributes for an app interface.
    '''

    # * attribute: module_path
    module_path: str

    # * attribute: class_name
    class_name: str

    # * attribute: attribute_id
    attribute_id: str


# ** contract: app_interface
class AppInterface(ModelContract):
    '''
    An app interface settings contract that defines the settings for an app interface.
    '''

    # * attribute: id
    id: str

    # * attribute: name
    name: str

    # * attribute: module_path
    module_path: str

    # * attribute: class_name
    class_name: str

    # * attribute: description
    description: str

    # * attribute: feature_flag
    feature_flag: str

    # * attribute: data_flag
    data_flag: str

    # * attribute: attributes
    attributes: List[AppAttribute]

    # * attribute: constants
    constants: Dict[str, Any]


# ** interface: app_repository
class AppRepository(Repository):
    '''
    An app repository is a class that is used to get an app interface.
    '''

    # * method: get_interface
    @abstractmethod
    def get_interface(self, app_name: str) -> AppInterface:
        '''
        Get the app interface settings by name.

        :param app_name: The name of the app. 
        :type app_name: str
        :return: The app interface.
        :rtype: AppInterface
        '''
        # Not implemented.
        raise NotImplementedError()
    
    # * method: list_interfaces
    @abstractmethod
    def list_interfaces(self) -> List[AppInterface]:
        '''
        List all app inferface settings.

        :return: A list of app settings.
        :rtype: List[AppInterface]
        '''
        # Not implemented.
        raise NotImplementedError()


# *** imports

# ** core
from typing import Dict, Any
import os
from importlib import import_module

# ** app
from .settings import *


# *** commands

# ** command: parse_parameter
class ParseParameter(Command):
    '''
    A command to parse a parameter from a string.
    '''

    # * method: execute
    def execute(self, parameter: str) -> Dict[str, Any]:
        '''
        Execute the command.

        :param parameter: The parameter to parse.
        :type parameter: str
        :return: The parsed parameter.
        :rtype: str
        '''

        # Parse the parameter.
        try:

            # If the parameter is an environment variable, get the value.
            if parameter.startswith('$env.'):
                result = os.getenv(parameter[5:])

                # Raise an exception if the environment variable is not found.
                if not result:
                    raise Exception('Environment variable not found.')

                # Return the result if the environment variable is found.
                return result

            # Return the parameter as is if it is not an environment variable.
            return parameter

        # Raise an error if the parameter parsing fails.
        except Exception as e:
            self.raise_error(
                'PARAMETER_PARSING_FAILED',
                f'Failed to parse parameter: {parameter}. Error: {str(e)}',
                parameter,
                str(e)
            )


# ** command: import_dependency
class ImportDependency(Command):
    '''
    A command to import a dependency from a module.
    '''

    # * method: execute
    def execute(self, module_path: str, class_name: str, **kwargs) -> Any:
        '''
        Execute the command.

        :param module_path: The module path to import from.
        :type module_path: str
        :param class_name: The class name to import.
        :type class_name: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The imported class instance.
        :rtype: Any
        '''

        # Import module.
        try:
            return getattr(import_module(module_path), class_name)

        # Raise an error if the dependency import fails.
        except Exception as e:
            self.raise_error(
                'IMPORT_DEPENDENCY_FAILED',
                f'Failed to import dependency: {module_path} from module {class_name}. Error: {str(e)}',
                module_path,
                class_name,
                str(e),
            )


# ** command: raise_error
class RaiseError(Command):
    '''
    A command to raise an error with a specific message.
    '''

    # * method: execute
    def execute(self, error_code: str, message: str = None, *args) -> None:
        '''
        Execute the command.

        :param error_code: The error code to raise.
        :type error_code: str
        :param args: Additional arguments for the error message.
        :type args: tuple
        '''

        # Raise an error with the specified code and arguments.
        raise TiferetError(error_code, message, *args)


# *** command_variables

# ** command_variable: parse_parameter
parse_parameter = ParseParameter()

# ** command_variable: import_dependency
import_dependency = ImportDependency()

# ** command_variable: raise_error
raise_error = RaiseError()



# *** imports 

# ** core
from typing import List, Any

# ** app
from .settings import *


# *** models

# ** model: error_message
class ErrorMessage(ValueObject):
    '''
    An error message object.
    '''

    # * attribute: lang
    lang = StringType(
        required=True,
        metadata=dict(
            description='The language of the error message text.'
        )
    )

    # * attribute: text
    text = StringType(
        required=True,
        metadata=dict(
            description='The error message text.'
        )
    )

    # * method: format
    def format(self, *args) -> str:
        '''
        Formats the error message text.

        :param args: The arguments to format the error message text with.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''

        # If there are no arguments, return the error message text.
        if not args:
            return self.text

        # Format the error message text and return it.
        return self.text.format(*args)


# ** model: error
class Error(Entity):
    '''
    An error object.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the error.'
        )
    )

    # * attribute: error_code
    error_code = StringType(
        metadata=dict(
            description='The unique code of the error.'
        )
    )

    # * attribute: message
    message = ListType(
        ModelType(ErrorMessage),
        required=True,
        metadata=dict(
            description='The error message translations for the error.'
        )
    )

    # * method: new
    @staticmethod
    def new(name: str, id: str = None, error_code: str = None, message: List[ErrorMessage | Any] = [], **kwargs) -> 'Error':
        '''Initializes a new Error object.

        :param name: The name of the error.
        :type name: str
        :param id: The unique identifier for the error.
        :type id: str
        :param error_code: The error code for the error.
        :type error_code: str
        :param message: The error message translations for the error.
        :type message: list 
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new Error object.
        '''
        
        # Set Id as the name lower cased if not provided.
        if not id:
            id = name.lower().replace(' ', '_')

        # Set the error code as the id upper cased if not provided.
        if not error_code:
            error_code = id.upper().replace(' ', '_')

        # Convert any error message dicts to ErrorMessage objects.
        message_objs = []
        for msg in message:
            if isinstance(msg, ErrorMessage):
                message_objs.append(msg)
            elif not isinstance(msg, ErrorMessage) and isinstance(msg, dict):
                message_objs.append(ValueObject.new(ErrorMessage, **msg))


        # Create and return a new Error object.
        return Entity.new(
            Error,
            id=id,
            name=name,
            error_code=error_code,
            message=message_objs,
            **kwargs
        )

    # * method: format_message
    def format_message(self, lang: str = 'en_US', *args) -> str:
        '''
        Formats the error message text for the specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param args: The format arguments for the error message text.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''

        # Iterate through the error messages.
        for msg in self.message:

            # Skip if the language does not match.
            if msg.lang != lang:
                continue

            # Format the error message text.
            return msg.format(*args)
        
    # * method: format_response
    def format_response(self, lang: str = 'en_US', *args, **kwargs) -> Any:
        '''
        Formats the error response for the specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param args: The format arguments for the error message text.
        :type args: tuple
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The formatted error response.
        :rtype: dict
        '''

        # Format the error message text.
        error_message = self.format_message(lang, *args)

        # If the error message is not found, return no response.
        if not error_message:
            return None

        # Return the formatted error response.
        return dict(
            error_code=self.error_code,
            message=error_message,
            **kwargs
        )
    
    # * method: set_message
    def set_message(self, lang: str, text: str):
        '''
        Sets the error message text for the specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param text: The error message text.
        :type text: str
        '''

        # Check if the message already exists for the language.
        for msg in self.message:
            if msg.lang == lang:
                msg.text = text
                return
        
        # If not, create a new ErrorMessage object and add it to the message list.
        self.message.append(
            ValueObject.new(
                ErrorMessage,
                lang=lang,
                text=text
            )
        )


# *** imports

# ** core
from typing import List, Dict, Tuple

# ** app
from ...data.container import ContainerAttributeYamlData
from ...contracts.container import ContainerRepository, ContainerAttribute
from ...clients import yaml as yaml_client


# *** proxies

# ** proxy: container_yaml_proxy
class ContainerYamlProxy(ContainerRepository):
    '''
    Yaml proxy for container attributes.
    '''

    # * init
    def __init__(self, container_config_file: str):
        '''
        Initialize the yaml proxy.
        
        :param container_config_file: The YAML file path for the container configuration.
        :type container_config_file: str
        '''

        # Set the container configuration file.
        self.config_file = container_config_file

    # * method: get_attribute
    def get_attribute(self, attribute_id: str) -> ContainerAttribute:
        '''
        Get the attribute from the yaml file.

        :param attribute_id: The attribute id.
        :type attribute_id: str
        :return: The container attribute.
        :rtype: ContainerAttribute
        '''

        # Load the attribute data from the yaml configuration file.
        data = yaml_client.load(
            self.config_file,
            create_data=lambda data: ContainerAttributeYamlData.from_data(
                id=attribute_id, **data),
            start_node=lambda data: data.get('attrs').get(attribute_id),
        )

        # If the data is None or the type does not match, return None.
        # Remove the type logic later, as the type parameter will be removed in v2 (obsolete).
        if data is None:
            return None
        
        # Return the attribute.
        return data.map()

    # * method: list_all
    def list_all(self) -> Tuple[List[ContainerAttribute], Dict[str, str]]:
        '''
        List all the container attributes and constants.

        :return: The list of container attributes and constants.
        :rtype: Tuple[List[ContainerAttribute], Dict[str, str]]
        '''

        # Define create data function to parse the YAML file.
        def create_data(data):
            
            # Create a list of ContainerAttributeYamlData objects from the YAML data.
            attrs = [
                ContainerAttributeYamlData.from_data(id=id, **attr_data)
                for id, attr_data in data.get('attrs', {}).items()
            ] if data.get('attrs') else []

            # Get the constants from the YAML data.
            consts = data.get('const', {}) if data.get('const') else {}

            # Return the parsed attributes and constants.
            return attrs, consts

        # Load the attribute data from the yaml configuration file.
        attr_data, consts = yaml_client.load(
            self.config_file,
            create_data=create_data
        )

        # Return the list of container attributes.
        return (
            [data.map() for data in attr_data],
            consts
        )



# *** imports

# ** app
from ..contracts.error import *


# *** handlers

# ** handler: error_handler
class ErrorHandler(ErrorService):
    '''
    A handler for error objects.
    '''

    # * attribute: error_repo
    error_repo: ErrorRepository

    # * method: init
    def __init__(self, error_repo: ErrorRepository):
        '''
        Initialize the error handler.

        :param error_repo: The error repository to use for handling errors.
        :type error_repo: ErrorRepository
        '''
        
        # Assign the attributes.
        self.error_repo = error_repo

    # * method: load_errors
    def load_errors(self, configured_errors: List[Error] = []) -> List[Error]:
        '''
        Load errors by their codes.

        :param configured_errors: The list of hard-coded errors to load.
        :type configured_errors: List[Error]
        :return: The list of loaded errors.
        :rtype: List[Error]
        '''
        
        # Load the errors from the repository.
        errors = self.error_repo.list()

        # If there are configured errors, extend the list with them.
        if configured_errors:
            errors.extend(configured_errors)

        # Return the list of errors.
        return errors


# *** imports

# ** app
from .settings import *
from ..models.app import *
from ..contracts.app import AppAttribute as AppAttributeContract
from ..contracts.app import AppInterface as AppInterfaceContract


# *** data

# ** data: app_attribute_yaml_data
class AppAttributeYamlData(AppAttribute, DataObject):
    '''
    A YAML data representation of an app dependency attribute object.
    '''

    # * attribute: attribute_id
    attribute_id = StringType(
        metadata=dict(
            description='The attribute id for the application dependency that is not required for assembly.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        serialized_name='params',
        deserialize_from=['params', 'parameters'],
        metadata=dict(
            description='The parameters for the application dependency that are not required for assembly.'
        ),
    )

    class Options():
        '''
        The options for the app dependency data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('parameters', 'attribute_id'),
            'to_data.yaml': DataObject.deny('attribute_id')
        }

    # * method: map
    def map(self, attribute_id: str, **kwargs) -> AppAttributeContract:
        '''
        Maps the app dependency data to an app dependency object.

        :param attribute_id: The id for the app dependency attribute.
        :type attribute_id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app attribute dependency contract.
        :rtype: AppAttributeContract
        '''

        # Map to the app dependency object.
        return super().map(
            AppAttribute,
            attribute_id=attribute_id,
            parameters=self.parameters,
            **self.to_primitive('to_model'),
            **kwargs
        )


# ** data: app_interface_yaml_data
class AppInterfaceYamlData(AppInterface, DataObject):
    '''
    A data representation of an app interface settings object.
    '''

    class Options():
        '''
        The options for the app interface data.
        '''
        serialize_when_none = False
        roles = {
            'to_model': DataObject.deny('attributes', 'constants', 'module_path', 'class_name'),
            'to_data.yaml': DataObject.deny('id')
        }

    # * attribute: module_path
    module_path = StringType(
        default=DEFAULT_MODULE_PATH,
        serialized_name='module',
        deserialize_from=['module_path', 'module'],
        metadata=dict(
            description='The app context module path for the app settings.'
        ),
    )

    # * attribute: class_name
    class_name = StringType(
        default=DEFAULT_CLASS_NAME,
        serialized_name='class',
        deserialize_from=['class_name', 'class'],
        metadata=dict(
            description='The class name for the app context.'
        ),
    )

    # * attribute: attributes
    attributes = DictType(
        ModelType(AppAttributeYamlData),
        default={},
        serialized_name='attrs',
        deserialize_from=['attrs', 'attributes'],
        metadata=dict(
            description='The app instance attributes.'
        ),
    )

    # * attribute: constants
    constants = DictType(
        StringType,
        default={},
        serialized_name='const',
        deserialize_from=['constants', 'const'],
        metadata=dict(
            description='The constants for the app settings.'
        ),
    )

    # * method: map
    def map(self, **kwargs) -> AppInterfaceContract:
        '''
        Maps the app interface data to an app interface contract.

        :param role: The role for the mapping.
        :type role: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new app interface contract.
        :rtype: AppInterfaceContract
        '''

        # Map the app interface data.
        return super().map(AppInterface,
            module_path=self.module_path,
            class_name=self.class_name,
            attributes=[attr.map(attribute_id=attr_id) for attr_id, attr in self.attributes.items()],
            constants=self.constants,
            **self.to_primitive('to_model'),
            **kwargs
        )


# *** imports

# ** core
from typing import Any, List

# ** app
from .cache import CacheContext
from ..handlers.container import ContainerService
from ..commands import *
from ..commands.dependencies import *


# *** contexts

# ** contexts: container_context
class ContainerContext(Model):
    '''
    A container context is a class that is used to create a container object.
    '''

    # * attribute: cache
    cache: CacheContext

    # * attribute: container_service
    container_service: ContainerService


    # * method: init
    def __init__(self, container_service: ContainerService, cache: CacheContext = None):
        '''
        Initialize the container context.

        :param container_service: The container service to use for executing container requests.
        :type container_service: ContainerService
        :param cache: The cache context to use for caching container data.
        :type cache: CacheContext
        '''

        # Assign the attributes.
        self.container_service = container_service
        self.cache = cache if cache else CacheContext()
    
    # * method: create_cache_key
    def create_cache_key(self, flags: List[str] = None) -> str:
        '''
        Create a cache key for the container.

        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: The cache key.
        :rtype: str
        '''

        # Create the cache key from the flags.
        return f"feature_container{'_' + '_'.join(flags) if flags else ''}"

    # * method: build_injector
    def build_injector(self,
            flags: List[str] = None,
        ) -> Injector:
        '''
        Build the container injector.

        :param flags: The feature or data flags to use.
        :type flags: List[str]
        :return: The container injector object.
        :rtype: Injector
        '''

        # Create the cache key for the injector from the flags.
        cache_key = self.create_cache_key(flags)

        # Check if the injector is already cached.
        cached_injector = self.cache.get(cache_key)
        if cached_injector:
            return cached_injector

        # Get all attributes and constants from the container service.
        attributes, constants = self.container_service.list_all()

        # Load constants from the attributes.
        constants = self.container_service.load_constants(attributes, constants, flags)

        # Create the dependencies for the injector.
        dependencies = {}
        for attr in attributes:
            try:
                dependencies[attr.id] = self.container_service.get_dependency_type(attr, flags)
            except TiferetError as e:
                raise e
                    

        # Create the injector with the dependencies and constants.
        injector = create_injector.execute(
            cache_key,
            dependencies=dependencies,
            **constants
        )

        # Cache the injector.
        self.cache.set(cache_key, injector)

        # Return the injector.
        return injector
    
    # * method: get_dependency
    def get_dependency(self, attribute_id: str, flags: List[str] = []) -> Any:
        '''
        Get an injector dependency by its attribute ID.

        :return: The container attribute.
        :rtype: Any
        '''

        # Get the cached injector.
        injector = self.build_injector(flags)

        # Get the dependency from the injector.
        dependency = get_dependency.execute(
            injector=injector,
            dependency_name=attribute_id,
        )
        
        # Raise an error if the dependency is not found.
        if not dependency:
            raise_error.execute(
                'CONTAINER_DEPENDENCY_NOT_FOUND',
                f'Dependency with ID {attribute_id} not found in container with flags {flags}.',
                attribute_id,
                flags
            )

        # Return the dependency.
        return dependency



# *** imports

# ** app
from .settings import *
from .core import import_dependency
from ..contracts.app import AppRepository, AppInterface


# *** commands:

# ** command: import_app_repository
class ImportAppRepository(Command):
    '''
    A command to import an app repository.
    '''

    # * method: execute
    def execute(self,
                app_repo_module_path: str = 'tiferet.proxies.yaml.app',
                app_repo_class_name: str = 'AppYamlProxy',
                app_repo_params: Dict[str, Any] = dict(
                    app_config_file='app/configs/app.yml'
                ),
                **kwargs
                ) -> AppRepository:
        '''
        Execute the command.

        :param app_repo_module_path: The application repository module path.
        :type app_repo_module_path: str
        :param app_repo_class_name: The application repository class name.
        :type app_repo_class_name: str
        :param app_repo_params: The application repository parameters.
        :type app_repo_params: dict
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The application repository instance.
        :rtype: AppRepository
        :raises TiferetError: If the import fails.
        '''

        # Import the app repository.
        try:
            result = import_dependency.execute(
                app_repo_module_path,
                app_repo_class_name
            )(**app_repo_params)

        # Raise an error if the import fails.
        except TiferetError as e:
            self.raise_error(
                'APP_REPOSITORY_IMPORT_FAILED',
                f'Failed to import app repository: {e}.',
                str(e)
            )

        # Return the imported app repository.
        return result


# ** command: get_app_interface
class GetAppInterface(Command):
    '''
    A command to get the application interface by its ID.
    '''

    def __init__(self, app_repo: AppRepository):
        '''
        Initialize the LoadAppInterface command.

        :param app_repo: The application repository instance.
        :type app_repo: AppRepository
        '''
        
        # Set the application repository.
        self.app_repo = app_repo

    # * method: execute
    def execute(self, interface_id: str, **kwargs) -> AppInterface:
        '''
        Execute the command to load the application interface.

        :param interface_id: The ID of the application interface to load.
        :type interface_id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The loaded application interface.
        :rtype: AppInterface
        :raises TiferetError: If the interface cannot be found.
        '''

        # Load the application interface.
        # Raise an error if the interface is not found.
        interface = self.app_repo.get_interface(interface_id)
        if not interface:
            self.raise_error(
                'APP_INTERFACE_NOT_FOUND',
                f'App interface with ID {interface_id} not found.'
            )

        # Return the loaded application interface.
        return interface


# *** imports

# ** core
import json


# *** classes


# ** class: tiferet_error
class TiferetError(Exception):
    '''
    A base exception for Tiferet.
    '''

    def __init__(self, error_code: str, message: str = None, *args):
        '''
        Initialize the exception.
        
        :param error_code: The error code.
        :type error_code: str
        :param message: An optional error message for internal exception handling.
        :type message: str
        :param args: Additional arguments for the error message.
        :type args: tuple
        '''

        # Set the error code and arguments.
        self.error_code = error_code

        # Initialize base exception with error data.
        super().__init__(
            json.dumps(dict(
                error_code=error_code,
                message=message,
            )), *args
        )


# *** imports

# ** app
from ..commands import *
from ..contracts.feature import *


# *** handlers

# ** handler: feature_handler
class FeatureHandler(FeatureService):
    '''
    Feature handler for executing feature requests.
    '''

    # * attribute: feature_repo
    feature_repo: FeatureRepository

    # * method: __init__
    def __init__(self, feature_repo: FeatureRepository):
        '''
        Initialize the feature handler.

        :param feature_repo: The feature repository to use for retrieving features.
        :type feature_repo: FeatureRepository
        '''

        # Assign the feature repository.
        self.feature_repo = feature_repo

    # * method: parse_parameter
    def parse_parameter(self, parameter: str, request: Request = None) -> str:
        '''
        Parse a parameter.

        
        :param parameter: The parameter to parse.
        :type parameter: str:
        param request: The request object containing data for parameter parsing.
        :type request: Request
        :return: The parsed parameter.
        :rtype: str
        '''

        # Parse the parameter if it not a request parameter.
        if not parameter.startswith('$r.'):
            return parse_parameter.execute(parameter)
        
        # Raise an error if the request is and the parameter comes from the request.
        if not request and parameter.startswith('$r.'):
            raise_error.execute(
                'REQUEST_NOT_FOUND',
                'Request data is not available for parameter parsing.',
                parameter
            )
    
        # Parse the parameter from the request if provided.
        result = request.data.get(parameter[3:], None)
        
        # Raise an error if the parameter is not found in the request data.
        if result is None:
            raise_error.execute(
                'PARAMETER_NOT_FOUND',
                f'Parameter {parameter} not found in request data.',
                parameter
            )

        # Return the parsed parameter.
        return result

    # * method: get_feature
    def get_feature(self, feature_id: str) -> Feature:
        '''
        Get a feature by its ID.

        :param feature_id: The ID of the feature to retrieve.
        :type feature_id: str
        :return: The feature model contract.
        :rtype: Feature
        '''

        # Retrieve the feature from the repository.
        feature = self.feature_repo.get(feature_id)

        # Verify the feature is not None.
        if not feature:
            raise_error.execute(
                'FEATURE_NOT_FOUND',
                f'Feature not found: {feature_id}',
                feature_id
            )

        # Return the feature.
        return feature


# *** imports

# ** app
from ...data import DataObject
from ...data.app import AppInterfaceYamlData
from ...contracts.app import AppRepository, AppInterface
from ...clients import yaml_client


# *** proxies

# ** proxy: app_yaml_proxy
class AppYamlProxy(AppRepository):

    # * field: config_file
    config_file: str = None

    # * method: init
    def __init__(self, app_config_file: str):
        '''
        Initialize the YAML proxy.

        :param app_config_file: The application configuration file.
        :type app_config_file: str
        '''

        # Set the configuration file.
        self.config_file = app_config_file

    # * method: list_interfaces
    def list_interfaces(self) -> list[AppInterface]:
        '''
        List all app interfaces.

        :return: The list of app interfaces.
        :rtype: List[AppInterface]
        '''

        # Load the app interface data from the yaml configuration file and map it to the app interface object.
        interfaces = yaml_client.load(
            self.config_file,
            create_data=lambda data: [
                DataObject.from_data(
                    AppInterfaceYamlData,
                    id=interface_id,
                    **record
                ).map() for interface_id, record in data.items()],
            start_node=lambda data: data.get('interfaces'))

        # Return the list of app interface objects.
        return interfaces

    # * method: get_interface
    def get_interface(self, id: str) -> AppInterface:
        '''
        Get the app interface.

        :param id: The app interface id.
        :type id: str
        :return: The app interface.
        :rtype: AppInterface
        '''

        # Load the app interface data from the yaml configuration file.
        _data: AppInterface = yaml_client.load(
            self.config_file,
            create_data=lambda data: DataObject.from_data(
                AppInterfaceYamlData,
                id=id, 
                **data
            ),
            start_node=lambda data: data.get('interfaces').get(id))

        # Return the app interface object.
        # If the data is None, return None.
        try:
            return _data.map()
        except AttributeError:
            return None


# *** imports

# ** core
from typing import Any

# ** infra
from schematics import Model

# ** app
from ..configs import *


# *** models

# ** model: model_object
class ModelObject(Model):
    '''
    A domain model object.
    '''

    # * method: new
    @staticmethod
    def new(
        model_type: type,
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> Any:
        '''
        Initializes a new model object.

        :param model_type: The type of model object to create.
        :type model_type: type
        :param validate: True to validate the model object.
        :type validate: bool
        :param strict: True to enforce strict mode for the model object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new model object.
        :rtype: Any
        '''

        # Create a new model object.
        _object = model_type(dict(
            **kwargs
        ), strict=strict)

        # Validate if specified.
        if validate:
            _object.validate()

        # Return the new model object.
        return _object


# ** model: entity
class Entity(ModelObject):
    '''
    A domain model entity.
    '''

    # ** attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The entity unique identifier.'
        )
    )


# ** model: value_object
class ValueObject(ModelObject):
    '''
    A domain model value object.
    '''

    pass



# ** model: module_dependency
class ModuleDependency(ValueObject):
    '''
    A module dependency.
    '''

    # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path.'
        )
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name.'
        )
    )



# *** imports

# ** app
from .settings import *


# *** models

# ** model: app_attribute
class AppAttribute(ValueObject):
    
    # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path for the app dependency.'
        )
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name for the app dependency.'
        )
    )

    # * attribute: attribute_id
    attribute_id = StringType(
        required=True,
        metadata=dict(
            description='The attribute id for the application dependency.'
        ),
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The parameters for the application dependency.'
        ),
    )


# ** model: app_interface
class AppInterface(Entity):
    '''
    The base application interface object.
    '''

    # * attribute: name
    name = StringType(
        required=True,
        metadata=dict(
            description='The name of the application interface.'
        ),
    )  

    # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path for the application instance context.'
        ),
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name for the application instance context.'
        ),
    )

    # attribute: feature_flag
    feature_flag = StringType(
        default='default',
        metadata=dict(
            description='The feature flag.'
        ),
    )

    # attribute: data_flag
    data_flag = StringType(
        default='default',
        metadata=dict(
            description='The data flag.'
        ),
    )

    # * attribute: attributes
    attributes = ListType(
        ModelType(AppAttribute),
        required=True,
        default=[],
        metadata=dict(
            description='The application instance attributes.'
        ),
    )

    # * attribute: constants
    constants = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The application dependency constants.'
        ),
    )

    # * method: add_attribute
    def add_attribute(self, module_path: str, class_name: str, attribute_id: str):
        '''
        Add a dependency attribute to the app interface.

        :param module_path: The module path for the app dependency attribute.
        :type module_path: str
        :param class_name: The class name for the app dependency attribute.
        :type class_name: str
        :param attribute_id: The id for the app dependency attribute.
        :type attribute_id: str
        :return: The added dependency.
        :rtype: AppDependency
        '''

        # Create a new AppDependency object.
        dependency = ValueObject.new(
            AppAttribute,
            module_path=module_path,
            class_name=class_name,
            attribute_id=attribute_id
        )

        # Add the dependency to the list of dependencies.
        self.attributes.append(dependency)
    
    # * method: get_attribute
    def get_attribute(self, attribute_id: str) -> AppAttribute:
        '''
        Get the dependency attribute by attribute id.

        :param attribute_id: The attribute id of the dependency attribute.
        :type attribute_id: str
        :return: The dependency attribute.
        :rtype: AppAttribute
        '''

        # Get the dependency attribute by attribute id.
        return next((attr for attr in self.attributes if attr.attribute_id == attribute_id), None)
    


# *** imports

# ** core
from abc import abstractmethod
from typing import Any

# ** app
from .settings import Service


# ** contracts

# ** contract: cache_service
class CacheService(Service):
    '''
    A contract for cache services in Tiferet applications.
    '''

    # * method: get
    @abstractmethod
    def get(self, key: str) -> Any:
        '''
        Retrieve an item from the cache.

        :param key: The key of the item to retrieve.
        :type key: str
        :return: The cached item or None if not found.
        :rtype: Any
        '''
        raise NotImplementedError(
            'The get method must be implemented by the cache service.'
        )

    # * method: set
    @abstractmethod
    def set(self, key: str, value: Any):
        '''
        Store an item in the cache.

        :param key: The key to store the value under.
        :type key: str
        :param value: The value to store.
        :type value: Any
        '''
        raise NotImplementedError(
            'The set method must be implemented by the cache service.'
        )

    # * method: delete
    @abstractmethod
    def delete(self, key: str):
        '''
        Delete an item from the cache.

        :param key: The key of the item to delete.
        :type key: str
        '''
        raise NotImplementedError(
            'The delete method must be implemented by the cache service.'
        )

    # * method: clear
    @abstractmethod
    def clear(self):
        '''
        Clear the entire cache.
        '''
        raise NotImplementedError(
            'The clear method must be implemented by the cache service.'
        )


# *** imports

# ** core
from typing import List, Dict, Any
from abc import abstractmethod

# ** app
from .settings import *


# *** contracts

# ** contract: error_message
class ErrorMessage(ModelContract):
    '''
    Contract for an error message translation.
    '''

    # * attribute: lang
    lang: str

    # * attribute: text
    text: str

    # * method: format
    @abstractmethod
    def format(self, *args) -> str:
        '''
        Format the error message text with provided arguments.

        :param args: The arguments to format the error message text.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''
        raise NotImplementedError('The format method must be implemented by the error message.')


# ** contract: error
class Error(ModelContract):
    '''
    Contract for an error object with multilingual messages.
    '''

    # * attribute: id
    id: str

    # * attribute: name
    name: str

    # * attribute: error_code
    error_code: str

    # * attribute: message
    message: List[ErrorMessage]

    # * method: format_message
    @abstractmethod
    def format_message(self, lang: str = 'en_US', *args) -> str:
        '''
        Format the error message for a specified language.

        :param lang: The language of the error message text (default: en_US).
        :type lang: str
        :param args: The format arguments for the error message text.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''
        raise NotImplementedError('The format_message method must be implemented by the error.')

    # * method: format_response
    @abstractmethod
    def format_response(self, lang: str = 'en_US', *args, **kwargs) -> Dict[str, Any]:
        '''
        Generate a formatted error response for a specified language.

        :param lang: The language of the error message text (default: en_US).
        :type lang: str
        :param args: The format arguments for the error message text.
        :type args: tuple
        :param kwargs: Additional keyword arguments for the response.
        :type kwargs: dict
        :return: The formatted error response.
        :rtype: Dict[str, Any]
        '''
        raise NotImplementedError('The format_response method must be implemented by the error.')

    # * method: set_message
    @abstractmethod
    def set_message(self, lang: str, text: str) -> None:
        '''
        Set or update the error message text for a specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param text: The error message text.
        :type text: str
        '''
        raise NotImplementedError('The set_message method must be implemented by the error.')


# ** contract: error_repository
class ErrorRepository(Repository):
    '''
    Contract for an error repository to manage error objects.
    '''

    # * method: exists
    @abstractmethod
    def exists(self, id: str, **kwargs) -> bool:
        '''
        Check if the error exists.

        :param id: The error id.
        :type id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: Whether the error exists.
        :rtype: bool
        '''
        raise NotImplementedError('The exists method must be implemented by the error repository.')

    # * method: get
    @abstractmethod
    def get(self, id: str) -> Error:
        '''
        Get an error object by its ID.

        :param id: The error id.
        :type id: str
        :return: The error object.
        :rtype: Error
        '''
        raise NotImplementedError('The get method must be implemented by the error repository.')

    # * method: list
    @abstractmethod
    def list(self) -> List[Error]:
        '''
        List all error objects.

        :return: The list of error objects.
        :rtype: List[Error]
        '''
        raise NotImplementedError('The list method must be implemented by the error repository.')

    # * method: save
    @abstractmethod
    def save(self, error: Error) -> None:
        '''
        Save the error.

        :param error: The error.
        :type error: Error
        '''
        raise NotImplementedError('The save method must be implemented by the error repository.')
    

# ** contract: error_service
class ErrorService(Service):
    '''
    Contract for an error service to handle error operations.
    '''

    # * method: load_errors
    @abstractmethod
    def load_errors(self, configured_errors: List[Error] = []) -> List[Error]:
        '''
        Load errors by their codes.

        :param configured_errors: The list of hard-coded errors to load.
        :type configured_errors: List[Error]
        :return: The list of loaded errors.
        :rtype: List[Error]
        '''
        raise NotImplementedError('The load_errors method must be implemented by the error service.')



# *** imports

# ** infra
import yaml


# *** functions

# ** function: load
def load(path: str, create_data = lambda data: data, start_node = lambda data: data, **kwargs):
    '''
    Load the data from the yaml file.

    :param path: The path to the yaml file.
    :type path: str
    :param create_data: The function to create the data.
    :type create_data: function
    :param start_node: The function to start the node.
    :type start_node: function
    :param kwargs: Additional keyword arguments.
    :type kwargs: dict
    :return: The data.
    :rtype: dict
    '''

    # Load the data from the yaml file.
    with open(path, 'r') as file:
        data = yaml.safe_load(file)

    # Find the start node.
    try:
        data = start_node(data)
    except AttributeError:
        return None
    
    # Return the data if it is None.
    if data == None:
        return None
    
    # Create and return the data.
    return create_data(data, **kwargs)


# ** function: save
def save(yaml_file: str, data: dict, data_save_path: str = None):
    '''
    Save the data to the yaml file.

    :param yaml_file: The path to the yaml file.
    :type yaml_file: str
    :param data: The data to save.
    :type data: dict
    :param data_save_path: The path to save the data to.
    :type data_save_path: str
    '''

    # Save the data to the yaml file and exit if no save path is provided.
    if not data_save_path:
        with open(yaml_file, 'w') as file:
            yaml.safe_dump(data, file)
        return

    # Load the yaml data.
    with open(yaml_file, 'r') as file:
        yaml_data = yaml.safe_load(file)

    # Get the data save path list.
    save_path_list = data_save_path.split('/')

    # Update the yaml data.
    new_yaml_data = None
    for fragment in save_path_list[:-1]:

        # If the new yaml data exists, update it.
        try:
            new_yaml_data = new_yaml_data[fragment]

        # If the new yaml data does not exist, create it from the yaml data.
        except TypeError:
            try:
                new_yaml_data = yaml_data[fragment]
                continue  
        
            # If the fragment does not exist, create it.
            except KeyError:
                new_yaml_data = yaml_data[fragment] = {}

        # If the fragment does not exist, create it.
        except KeyError: 
            new_yaml_data[fragment] = {}
            new_yaml_data = new_yaml_data[fragment]

    # Update the yaml data.
    try:
        new_yaml_data[save_path_list[-1]] = data
    # if there is a type error because the new yaml data is None, update the yaml data directly.
    except TypeError:
        yaml_data[save_path_list[-1]] = data

    # Save the updated yaml data.
    with open(yaml_file, 'w') as file:
        yaml.safe_dump(yaml_data, file)



# *** imports

# ** app
from .container import ContainerContext
from .cache import CacheContext
from ..handlers.feature import FeatureService
from ..models.feature import Request
from ..commands import *


# *** contexts

# ** context: feature_context
class FeatureContext(object):

    # * attribute: container
    container: ContainerContext

    # * attribute: cache
    cache: CacheContext

    # * attribute: feature_service
    feature_service: FeatureService

    # * method: init
    def __init__(self, 
            feature_service: FeatureService,
            container: ContainerContext,
            cache: CacheContext = None):
        '''
        Initialize the feature context.

        :param feature_service: The feature service to use for executing feature requests.
        :type feature_service: FeatureService
        :param container: The container context for dependency injection.
        :type container: ContainerContext
        :param cache: The cache context to use for caching feature data.
        :type cache: CacheContext
        '''

        # Assign the attributes.
        self.feature_service = feature_service
        self.container = container
        self.cache = cache if cache else CacheContext()

    # * method: load_feature_command
    def load_feature_command(self, attribute_id: str) -> Command:
        '''
        Load a feature command by its attribute ID from the container.

        :param attribute_id: The attribute ID of the command to load.
        :type attribute_id: str
        :return: The command object.
        :rtype: Command
        '''

        # Attempt to retrieve the command from the container.
        try:
            return self.container.get_dependency(attribute_id)
        
        # If the command is not found, raise an error.
        except Exception as e:
            raise_error.execute(
                'FEATURE_COMMAND_LOADING_FAILED',
                f'Failed to load feature command attribute: {attribute_id}. Ensure the container attributes is configured with the appropriate default settings/flags.',
                attribute_id,
                str(e)
        )
            
    # * method: handle_command
    def handle_command(self,
        command: Command, 
        request: Request,
        data_key: str = None,
        pass_on_error: bool = False,
        **kwargs
    ):
        '''
        Handle the execution of a command with the provided request and command-handling options.
        
        :param command: The command to execute.
        :type command: Command
        :param request: The request context object.
        :type request: Request
        :param debug: Debug flag.
        :type debug: bool
        :param data_key: Optional key to store the result in the request data.
        :type data_key: str
        :param pass_on_error: If True, pass on the error instead of raising it.
        :type pass_on_error: bool
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Execute the command with the request data and parameters, and optional contexts.
        try:
            result = command.execute(
                **request.data,
                **kwargs
            )
        except Exception as e:
            if not pass_on_error:
                raise e

        # If a data key is provided, store the result in the request data.
        if data_key:
            request.data[data_key] = result

        # Otherwise, set the result.
        else:
            request.set_result(result)

        
    # * method: execute_feature
    def execute_feature(self, feature_id: str, request: Request, **kwargs):
        '''
        Execute a feature by its ID with the provided request.
        
        :param request: The request object.
        :type request: Request
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        '''

        # Try to get the feature by its id from the cache.
        # If it does not exist, retrieve it from the feature handler and cache it.
        feature = self.cache.get(feature_id)
        if not feature:
            feature = self.feature_service.get_feature(feature_id)
            self.cache.set(feature_id, feature)

        # Load the command dependencies from the container for the feature.
        commands = [
            self.load_feature_command(cmd.attribute_id)
            for cmd in feature.commands
        ]
              
        # Execute the feature with the request and commands.
        for index, cmd in enumerate(commands):

            # Get the feature command from the feature.
            feature_command = feature.commands[index]

            # Parse the command parameters
            params = {
                param: self.feature_service.parse_parameter(value, request)
                for param, value in feature_command.parameters.items()
            }

            # Execute the command with the request data and parameters.
            self.handle_command(
                cmd,
                request,
                data_key=feature_command.data_key,
                pass_on_error=feature_command.pass_on_error,
                **params,
                features=self.feature_service,
                container=self.container,
                cache=self.cache,
                **kwargs
            )
            


# *** imports

# ** core
from typing import Dict, Any

# ** app
from ..configs import TiferetError


# *** classes

# ** class: command
class Command(object):
    '''
    A base class for an app command object.
    '''

    # * method: execute
    def execute(self, **kwargs) -> Any:
        '''
        Execute the service command.
        
        :param kwargs: The command arguments.
        :type kwargs: dict
        :return: The command result.
        :rtype: Any
        '''

        # Not implemented.
        raise NotImplementedError()

    # * method: raise_error
    def raise_error(self, error_code: str, message: str = None, *args):
        '''
        Raise an error with the given error code and arguments.

        :param error_code: The error code.
        :type error_code: str
        :param message: The error message.
        :type message: str
        :param args: Additional error arguments.
        :type args: tuple
        '''

        # Raise the TiferetError with the given error code and arguments.
        raise TiferetError(
            error_code,
            message,
            *args
        )    

    # * method: verify
    def verify(self, expression: bool, error_code: str, message: str = None, *args):
        '''
        Verify an expression and raise an error if it is false.

        :param expression: The expression to verify.
        :type expression: bool
        :param error_code: The error code.
        :type error_code: str
        :param message: The error message.
        :type message: str
        :param args: Additional error arguments.
        :type args: tuple
        '''

        # Verify the expression.
        try:
            assert expression
        except AssertionError:
            self.raise_error(
                error_code,
                message,
                *args
            )

    # * method: handle
    @staticmethod
    def handle(
            command: type,
            dependencies: Dict[str, Any] = {},
            **kwargs) -> Any:
        '''
        Handle an app command instance.

        :param command: The command to handle.
        :type command: type
        :param dependencies: The command dependencies.
        :type dependencies: Dict[str, Any]
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: The result of the command.
        :rtype: Any
        '''

        # Get the command handler.
        command_handler = command(**dependencies)

        # Execute the command handler.
        result = command_handler.execute(**kwargs)
        return result


# *** imports

# ** core
from typing import List, Dict, Any
from abc import abstractmethod

# ** app
from .settings import *


# *** contacts

# ** contract: request
class Request(ModelContract):
    '''
    Request contract for feature execution.
    '''

    # * attribute: headers
    headers: Dict[str, str]

    # * attribute: data
    data: Dict[str, Any]

    # * attribute: debug
    debug: bool

    # * attribute: result
    result: str

    # * method: set_result
    def set_result(self, result: Any):
        '''
        Set the result of the request.

        :param result: The result to set.
        :type result: Any
        '''
        raise NotImplementedError('The set_result method must be implemented by the request model.')


# ** contract: feature_command
class FeatureCommand(ModelContract):
    '''
    Feature command contract.
    '''

    # * attribute: id
    id: str

    # * attribute: name
    name: str

    # * attribute: description
    description: str

    # * attribute: data_key
    data_key: str

    # * attribute: pass_on_error
    pass_on_error: bool

    # * attribute: parameters
    parameters: Dict[str, Any]


# ** contract: feature
class Feature(ModelContract):
    '''
    Feature contract.
    '''

    # * attribute: id
    id: str

    # * attribute: commands
    commands: List[FeatureCommand]


# ** contract: feature_repository
class FeatureRepository(Repository):
    '''
    Feature repository interface.
    '''

    # * method: exists
    @abstractmethod
    def exists(self, id: str) -> bool:
        '''
        Verifies if the feature exists.

        :param id: The feature id.
        :type id: str
        :return: Whether the feature exists.
        :rtype: bool
        '''
        raise NotImplementedError('The exists method must be implemented by the feature repository.')

    # * method: get
    @abstractmethod
    def get(self, id: str) -> Feature:
        '''
        Get the feature by id.

        :param id: The feature id.
        :type id: str
        :return: The feature object.
        :rtype: Any
        '''
        raise NotImplementedError('The get method must be implemented by the feature repository.')

    # * method: list
    @abstractmethod
    def list(self, group_id: str = None) -> List[Feature]:
        '''
        List the features.

        :param group_id: The group id.
        :type group_id: str
        :return: The list of features.
        :rtype: List[Feature]
        '''
        raise NotImplementedError('The list method must be implemented by the feature repository.')
    


# ** contract: feature_service
class FeatureService(Service):
    '''
    Feature service contract.
    '''

    # * method: parse_parameter
    @abstractmethod
    def parse_parameter(self, parameter: str, request: Request = None) -> str:
        '''
        Parse a parameter.

        :param parameter: The parameter to parse.
        :type parameter: str
        :param request: The request object containing data for parameter parsing.
        :type request: Request
        :return: The parsed parameter.
        :rtype : str
        '''
        raise NotImplementedError('The parse_parameter method must be implemented by the feature service.')

    # * method: get_feature
    @abstractmethod
    def get_feature(self, feature_id: str) -> Feature:
        '''
        Get a feature by its ID.

        :param feature_id: The ID of the feature to retrieve.
        :type feature_id: str
        :return: The feature object.
        :rtype: Feature
        '''
        raise NotImplementedError('The get_feature method must be implemented by the feature service.')



# *** imports

# ** infra
from abc import ABC, abstractmethod


# *** classes


# ** class: model_contract
class ModelContract(ABC):
    '''
    The model contract interface as an abstract base class.
    '''

    pass


# ** class: repository
class Repository(ABC):
    '''
    The repository interface as an abstract base class.
    '''

    pass


# ** class: service
class Service(ABC):
    '''
    The service interface as an abstract base class.
    '''

    pass


# *** imports

# ** core
from typing import List

# ** app
from ...clients import yaml_client
from ...contracts.feature import Feature, FeatureRepository
from ...data import DataObject
from ...data.feature import FeatureData as FeatureYamlData


# *** proxies

# ** proxies: feature_yaml_proxy
class FeatureYamlProxy(FeatureRepository):
    '''
    Yaml repository for features.
    '''

    # * method: init
    def __init__(self, feature_config_file: str):
        '''
        Initialize the yaml repository.

        :param feature_config_file: The feature configuration file.
        :type feature_config_file: str
        '''

        # Set the base path.
        self.config_file = feature_config_file

    # * method: exists
    def exists(self, id: str) -> bool:
        '''
        Verifies if the feature exists.
        
        :param id: The feature id.
        :type id: str
        :return: Whether the feature exists.
        :rtype: bool
        '''

        # Retrieve the feature by id.
        feature = self.get(id)

        # Return whether the feature exists.
        return feature is not None

    # * method: get
    def get(self, id: str) -> Feature:
        '''
        Get the feature by id.
        
        :param id: The feature id.
        :type id: str
        :return: The feature object.
        '''

        # Get the feature.
        return next((feature for feature in self.list() if feature.id == id), None)
    
    # * method: list
    def list(self, group_id: str = None) -> List[Feature]:
        '''
        List the features.
        
        :param group_id: The group id.
        :type group_id: str
        :return: The list of features.
        :rtype: List[Feature]
        '''

        # Load all feature data from yaml.
        features = yaml_client.load(
            self.config_file,
            create_data=lambda data: [DataObject.from_data(
                FeatureYamlData,
                id=id,
                feature_key=id.split('.')[-1],
                group_id=id.split('.')[0] if not group_id else group_id,
                **feature_data
            ) for id, feature_data in data.items()],
            start_node=lambda data: data.get('features')
        )

        # Filter features by group id.
        if group_id:
            features = [feature for feature in features if feature.group_id == group_id]

        # Return the list of features.
        return [feature.map() for feature in features]


# *** imports

# ** core
from typing import List, Dict, Tuple, Any
from abc import abstractmethod

# ** app
from .settings import *


# *** contracts

# ** contract: flagged_dependency
class FlaggedDependency(ModelContract):
    '''
    A contract for flagged dependencies.
    '''

    # * attribute: flag
    flag: str

    # * attribute: parameters
    parameters: Dict[str, str]

    # * attribute: module_path
    module_path: str

    # * attribute: class_name
    class_name: str


# ** contract: container_attribute
class ContainerAttribute(ModelContract):
    '''
    A contract defining container injector attributes.
    '''

    # * attribute: id
    id: str

    # * attribute: module_path
    module_path: str

    # * attribute: class_name
    class_name: str

    # * attribute: parameters
    parameters: Dict[str, Any]

    # * attribute: dependencies
    dependencies: List[FlaggedDependency]

    # * method: get_dependency
    @abstractmethod
    def get_dependency(self, *flags) -> FlaggedDependency:
        '''
        Gets a container dependency by flag.

        :param flags: The flags for the flagged container dependency.
        :type flags: Tuple[str, ...]
        :return: The container dependency.
        :rtype: FlaggedDependency
        '''
        raise NotImplementedError('get_dependency method must be implemented in the ContainerAttribute class.')


# ** contract: container_repository
class ContainerRepository(Repository):
    '''
    An interface for accessing container attributes.
    '''

    # * method: get_attribute
    @abstractmethod
    def get_attribute(self, attribute_id: str, flag: str = None) -> ContainerAttribute:
        '''
        Get the attribute from the container repository.

        :param attribute_id: The attribute id.
        :type attribute_id: str
        :param flag: An optional flag to filter the attribute.
        :type flag: str
        :return: The container attribute.
        :rtype: ContainerAttribute
        '''
        raise NotImplementedError('get_attribute method must be implemented in the ContainerRepository class.')

    # * method: list_all
    @abstractmethod
    def list_all(self) -> Tuple[List[ContainerAttribute], Dict[str, str]]:
        '''
        List all the container attributes and constants.

        :return: The list of container attributes and constants.
        :rtype: Tuple[List[ContainerAttribute], Dict[str, str]]
        '''
        raise NotImplementedError('list_all method must be implemented in the ContainerRepository class.')


# ** contract: container_service
class ContainerService(Service):
    '''
    An interface for accessing container dependencies.
    '''

   # * method: list_all
    @abstractmethod
    def list_all(self) -> Tuple[List[ContainerAttribute], Dict[str, str]]:
        '''
        List all container attributes and constants.

        :return: A tuple containing a list of container attributes and a dictionary of constants.
        :rtype: Tuple[List[ContainerAttribute], Dict[str, str]]
        '''
        raise NotImplementedError('list_all method must be implemented in the ContainerService class.')
    
     # * method: load_constants
    @abstractmethod
    def load_constants(self, attributes: List[ContainerAttribute], constants: Dict[str, str] = {}, flags: List[str] = []) -> Dict[str, str]:
        '''
        Load constants from the container attributes.

        :param attributes: The list of container attributes.
        :type attributes: List[ContainerAttribute]
        :param constants: The dictionary of constants.
        :type constants: Dict[str, str]
        :param flags: Optional list of flags to filter the constants.
        :type flags: List[str]
        :return: A dictionary of constants.
        :rtype: Dict[str, str]
        '''
        raise NotImplementedError('load_constants method must be implemented in the ContainerService class.')
    
    # * method: get_dependency_type
    @abstractmethod
    def get_dependency_type(self, attribute: ContainerAttribute, flags: List[str] = []) -> type:
        '''
        Get the type of a container attribute.

        :param attribute: The container attribute.
        :type attribute: ContainerAttribute
        :param flags: Optional list of flags to filter the dependency type.
        :type flags: List[str]
        :return: The type of the container attribute.
        :rtype: type
        '''
        raise NotImplementedError('get_dependency_type method must be implemented in the ContainerService class.')


# *** imports


# *** configs

# ** config: errors
ERRORS = [
    dict(
        id='parameter_parsing_failed',
        name='Parameter Parsing Failed',
        error_code='PARAMETER_PARSING_FAILED',
        message=[
            dict(lang='en_US', text='Failed to parse parameter: {}. Error: {}')
        ]
    ),
    dict(
        id='import_dependency_failed',
        name='Import Dependency Failed',
        error_code='IMPORT_DEPENDENCY_FAILED',
        message=[
            dict(lang='en_US', text='Failed to import dependency: {} from module {}. Error: {}')
        ]
    ),
    dict(
        id='app_repository_import_failed',
        name='App Repository Import Failed',
        error_code='APP_REPOSITORY_IMPORT_FAILED',
        message=[
            dict(lang='en_US', text='Failed to import app repository: {}.')
        ]
    ),
    dict(
        id='app_interface_not_found',
        name='App Interface Not Found',
        error_code='APP_INTERFACE_NOT_FOUND',
        message=[
            dict(lang='en_US', text='App interface with ID {} not found.')
        ]
    ),
    dict(
        id='feature_command_loading_failed',
        name='Feature Command Loading Failed',
        error_code='FEATURE_COMMAND_LOADING_FAILED',
        message=[
            dict(lang='en_US', text='Failed to load feature command attribute: {}. Ensure the container attributes are configured with the appropriate default settings/flags. {}')
        ]
    )
]


# *** imports

# ** core
from typing import Any

# ** app
from ..contracts.settings import ModelContract
from ..models.settings import *


# *** constants

# ** constant: default_module_path
DEFAULT_MODULE_PATH = 'tiferet.contexts.app'

# ** constant: default_class_name
DEFAULT_CLASS_NAME = 'AppInterfaceContext'


# *** classes

# ** class: data_object
class DataObject(Model):
    '''
    A data representation object.
    '''

    # ** method: map
    def map(self,
            type: ModelObject,
            role: str = 'to_model',
            validate: bool = True,
            **kwargs
            ) -> ModelContract:
        '''
        Maps the model data to a model object.

        :param type: The type of model object to map to.
        :type type: type
        :param role: The role for the mapping.
        :type role: str
        :param validate: True to validate the model object.
        :type validate: bool
        :param kwargs: Additional keyword arguments for mapping.
        :type kwargs: dict
        :return: A new model object as a contract.
        :rtype: ModelContract
        '''

        # Get primitive of the model data and merge with the keyword arguments.
        # Give priority to the keyword arguments.
        _data = self.to_primitive(role=role)
        for key, value in kwargs.items():
            _data[key] = value

        # Map the data object to a model object.
        # Attempt to create a new model object with a custom factory method.
        # If the factory method does not exist, employ the standard method.
        try:
            _object = type.new(**_data, strict=False)
        except Exception:
            _object = ModelObject.new(type, **_data, strict=False)

        # Validate if specified.
        if validate:
            _object.validate()

        # Return the model data.
        return _object

    # ** method: from_model
    @staticmethod
    def from_model(
        data: 'DataObject',
        model: ModelObject,
        validate: bool = True,
        **kwargs
    ) -> 'DataObject':
        '''
        Initializes a new data object from a model object.

        :param model: The type of model object to map from.
        :type model: type
        :param data: The data object to map from.
        :type data: DataObject
        :param validate: True to validate the data object.
        :type validate: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new data object.
        :rtype: DataObject
        '''

        # Create a new data object.
        obj = data(dict(
            **model.to_primitive(),
            **kwargs
        ), strict=False)

        # Validate the data object if specified.
        if validate:
            obj.validate()

        # Return the data object.
        return obj

    # ** method: from_data
    @staticmethod
    def from_data(
        data: type,
        **kwargs
    ) -> 'DataObject':
        '''
        Initializes a new data object from a dictionary.

        :param data: The type of data object to map from.
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new data object.
        :rtype: DataObject
        '''

        # Create a new data object.
        return data(dict(**kwargs), strict=False)

    # ** method: allow
    @staticmethod
    def allow(*args) -> Any:
        '''
        Creates a whitelist transform for data mapping.

        :param args: Fields to allow in the transform.
        :type args: tuple
        :return: The whitelist transform.
        :rtype: Any
        '''

        # Create a whitelist transform.
        # Create a wholelist transform if no arguments are specified.
        from schematics.transforms import whitelist, wholelist
        if args:
            return whitelist(*args)
        return wholelist()

    # ** method: deny
    @staticmethod
    def deny(*args) -> Any:
        '''
        Creates a blacklist transform for data mapping.

        :param args: Fields to deny in the transform.
        :type args: tuple
        :return: The blacklist transform.
        :rtype: Any
        '''

        # Create a blacklist transform.
        from schematics.transforms import blacklist
        return blacklist(*args)


# *** imports

# ** infra
from schematics import Model, types as t


# *** classes

# ** class: string_type
class StringType(t.StringType):
    '''
    A string type.
    '''

    pass


# ** class: integer_type
class IntegerType(t.IntType):
    '''
    An integer type.
    '''

    pass


# ** class: float_type
class FloatType(t.FloatType):
    '''
    A float type.
    '''

    pass


# ** class: boolean_type
class BooleanType(t.BooleanType):
    '''
    A boolean type.
    '''

    pass


# ** class: list_type
class ListType(t.ListType):
    '''
    A list type.
    '''

    pass


# ** class: dict_type
class DictType(t.DictType):
    '''
    A dictionary type.
    '''

    pass


# ** class: model_type
class ModelType(t.ModelType):
    '''
    A model type.
    '''

    pass


# ** class: model_object
class ModelObject(Model):
    '''
    A domain model object.
    '''

    # ** attribute: name
    name = StringType(
        metadata=dict(
            description='The name of the object.'
        )
    )

    # ** attribute: description
    description = StringType(
        metadata=dict(
            description='The description of the object.'
        )
    )

    # * method: new
    @staticmethod
    def new(
        model_type: type,
        validate: bool = True,
        strict: bool = True,
        **kwargs
    ) -> 'ModelObject':
        '''
        Initializes a new model object.

        :param model_type: The type of model object to create.
        :type model_type: type
        :param validate: True to validate the model object.
        :type validate: bool
        :param strict: True to enforce strict mode for the model object.
        :type strict: bool
        :param kwargs: Keyword arguments.
        :type kwargs: dict
        :return: A new model object.
        :rclass: ModelObject
        '''

        # Create a new model object.
        _object: ModelObject = model_type(dict(
            **kwargs
        ), strict=strict)

        # Validate if specified.
        if validate:
            _object.validate()

        # Return the new model object.
        return _object


# ** model: value_object
class ValueObject(ModelObject):
    '''
    A domain model value object.
    '''
    pass


# ** class: entity
class Entity(ModelObject):
    '''
    A domain model entity.
    '''

    # ** attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The entity unique identifier.'
        )
    )


# *** imports

# ** core
from typing import Any, Tuple, List

# ** app
from .cache import CacheContext
from ..configs.error import ERRORS
from ..handlers.error import ErrorService
from ..models.error import Error, ModelObject
from ..commands import raise_error, TiferetError


# *** contexts

# ** context: error_context
class ErrorContext(object):
    '''
    The error context object.
    '''

    # * attribute: error_service
    error_service: ErrorService

    # * attribute: cache
    cache: CacheContext

    # * method: init
    def __init__(self, error_service: ErrorService, cache: CacheContext = None):
        '''
        Initialize the error context.

        :param error_service: The error service to use for handling errors.
        :type error_service: ErrorService
        :param cache: The cache context to use for caching error data.
        :type cache: CacheContext
        '''

        # Assign the attributes.
        self.error_service = error_service
        self.cache = cache if cache else CacheContext()

    # * method: load_errors
    def load_errors(self) -> List[Error]:
        '''
        Load errors from the error service.

        :return: The list of loaded errors.
        :rtype: List[Error]
        '''

        # Convert configured errors to a list of error objects.
        configured_errors = [
            ModelObject.new(
                Error,
                **error_data
        ) for error_data in ERRORS]

        # Load errors from the error service, including any configured errors.
        return self.error_service.load_errors(configured_errors)

    
    # * method: get_error_by_code
    def get_error_by_code(self, error_code: str) -> Error:
        '''
        Get an error by its code.

        :param error_code: The error code to retrieve.
        :type error_code: str
        :return: The error object.
        :rtype: Error
        '''

        # Check to see if there is an error map in the cache.
        error_map = self.cache.get('error_map')

        # If there is no error map or the error code does not exist in the map,
        # we need to load the errors from the error service and create a new error map.
        # This is to ensure that we always have the latest errors available.
        if not error_map or not error_map.get(error_code):
            errors = self.load_errors()
            error_map = {error.error_code: error for error in errors}
            self.cache.set('error_map', error_map)

        # Raise an error if the error code does not exist in the error map.
        if not error_map or error_code not in error_map:
            raise_error.execute(
                'ERROR_NOT_FOUND', 
                f'Error not found: {error_code}.',
                error_code
            )

        # Return the error object from the error map.
        return error_map.get(error_code)

    # * method: handle_error
    def handle_error(self, exception: TiferetError, lang: str = 'en_US', **kwargs) -> Tuple[bool, Any]:
        '''
        Handle an error.

        :param exception: The exception to handle.
        :type exception: Exception
        :param lang: The language to use for the error message.
        :type lang: str
        :return: Whether the error was handled.
        :rtype: bool
        '''

        # Raise the exception if it is not a Tiferet error.
        if not isinstance(exception, TiferetError):
            raise exception

        # Get the error by its code from the error service.
        error = self.get_error_by_code(exception.error_code)
        
        # Format the error response.
        error_response = error.format_response(
            lang,
            *exception.args[1:],
            **kwargs
        )

        # Return the error response.
        return error_response



# *** imports

# ** app
from ..domain import *

# *** constants

# */ list[str]
CONTAINER_ATTRIBUTE_TYPE_CHOICES = [
    'interface',
    'feature',
    'data'
]


# *** models

# ** model: container_depenedency
class ContainerDependency(ModuleDependency):
    '''
    A container dependency object.
    '''

    # * attribute: flag
    flag = StringType(
        required=True,
        metadata=dict(
            description='The flag for the container dependency.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The container dependency parameters.'
        )
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'ContainerDependency':
        '''
        Initializes a new ContainerDependency object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ContainerDependency object.
        :rtype: ContainerDependency
        '''

        # Create and return a new ContainerDependency object.
        return super(ContainerDependency, ContainerDependency).new(
            ContainerDependency,
            **kwargs)


# ** model: container_attribute
class ContainerAttribute(Entity):
    '''
    An attribute that defines container injectior behavior.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier for the container attribute.'
        )
    )

    # * attribute: type
    type = StringType(
        required=True,
        choices=CONTAINER_ATTRIBUTE_TYPE_CHOICES,
        metadata=dict(
            description='The type of container attribute.'
        )
    )

    # * attribute: dependencies
    dependencies = ListType(
        ModelType(ContainerDependency),
        default=[],
        metadata=dict(
            description='The container attribute dependencies.'
        )
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'ContainerAttribute':
        '''
        Initializes a new ContainerAttribute object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ContainerAttribute object.
        :rtype: ContainerAttribute
        '''

        # Create and return a new ContainerAttribute object.
        return super(ContainerAttribute, ContainerAttribute).new(
            ContainerAttribute,
            **kwargs)
        
    # * method: get_dependency
    def get_dependency(self, flag: str) -> ContainerDependency:
        '''
        Gets a container dependency by flag.

        :param flag: The flag for the container dependency.
        :type flag: str
        :return: The container dependency.
        :rtype: ContainerDependency
        '''

        # Return the dependency with the matching flag.
        return next(
            (dependency for dependency in self.dependencies if dependency.flag == flag),
            None
        )
        
    # * method: set_dependency
    def set_dependency(self, dependency: ContainerDependency):
        '''
        Sets a container dependency.

        :param dependency: The container dependency to set.
        :type dependency: ContainerDependency
        '''

        # Replace the value of the dependency if a dependency with the same flag exists.
        for index, _dependency in enumerate(self.dependencies):
            if _dependency.flag == dependency.flag:
                self.dependencies[index] = dependency
                return
        
        # Append the dependency otherwise.
        self.dependencies.append(dependency)



try:
    from setuptools import setup
except:
    from distutils.core import setup

config = {
    'description': 'A multi-purpose application framework embodying beauty in form.',
    'author': 'Andrew Shatz, Great Strength Systems',
    'url': r'https://github.com/greatstrength/app',
    'download_url': r'https://github.com/greatstrength/app',
    'author_email': 'andrew@greatstrength.me',
    'version': '1.0.0-beta.2',
    'license': 'BSD 3',
    'install_requires': [
        'schematics>=2.1.1',
        'pyyaml>=6.0.1',
        'dependencies>=7.7.0'
    ],
    'packages': [
        'tiferet',
        'tiferet.clients',
        'tiferet.commands',
        'tiferet.configs',
        'tiferet.contexts',
        'tiferet.contracts',
        'tiferet.data',
        'tiferet.handlers',
        'tiferet.models',
        'tiferet.proxies',
        'tiferet.proxies.yaml',
    ],    
    'scripts': [],
    'name': 'tiferet',
    'extras_require': {
        'test': ['pytest>=8.3.3', 'pytest_env>=1.1.5'],
    }
}

setup(**config)


# *** imports

# ** core
from typing import List, Dict

# ** app
from ...data.error import ErrorData
from ...contracts.error import (
    Error,
    ErrorRepository
)
from ...clients import yaml as yaml_client


# *** proxies

# ** proxy: yaml_proxy
class ErrorYamlProxy(ErrorRepository):
    '''
    The YAML proxy for the error repository
    '''

    # * field: config_file
    config_file: str

    # * method: init
    def __init__(self, error_config_file: str):
        '''
        Initialize the yaml proxy.

        :param error_config_file: The error configuration file.
        :type error_config_file: str
        '''

        # Set the base path.
        self.config_file = error_config_file

    # * method: exists
    def exists(self, id: str, **kwargs) -> bool:
        '''
        Check if the error exists.
        
        :param id: The error id.
        :type id: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: Whether the error exists.
        :rtype: bool
        '''

        # Load the error data from the yaml configuration file.
        data = self.get(id)

        # Return whether the error exists.
        return data is not None

    # * method: get
    def get(self, id: str) -> Error:
        '''
        Get the error.
        
        :param id: The error id.
        :type id: str
        :return: The error.
        :rtype: Error
        '''

        # Load the error data from the yaml configuration file.
        _data: ErrorData = yaml_client.load(
            self.config_file,
            create_data=lambda data: ErrorData.from_data(
                id=id, **data),
            start_node=lambda data: data.get('errors').get(id))

        # Return the error object.
        return _data.map() if _data else None
    
    # * method: list
    def list(self) -> List[Error]:
        '''
        List all errors.

        :return: The list of errors.
        :rtype: List[Error]
        '''

        # Load the error data from the yaml configuration file.
        _data: Dict[str, ErrorData] = yaml_client.load(
            self.config_file,
            create_data=lambda data: {id: ErrorData.from_data(
                id=id, **error_data) for id, error_data in data.items()},
            start_node=lambda data: data.get('errors'))

        # Return the error object.
        return [data.map() for data in _data.values()]

    # * method: save
    def save(self, error: Error):
        '''
        Save the error.

        :param error: The error.
        :type error: Error
        '''

        # Create updated error data.
        error_data = ErrorData.from_model(ErrorData, error)

        # Update the error data.
        yaml_client.save(
            yaml_file=self.config_file,
            data=error_data.to_primitive(),
            data_save_path=f'errors/{error.id}',
        )



# *** imports

# ** core
from typing import Dict, Any

# ** app
from ..configs import *


# *** contexts

# ** context: cache_context
class CacheContext(object):
    '''
    A context for managing cache operations within Tiferet applications.
    '''
    
    # * attribute: cache (private)
    _cache = Dict[str, Any]

    # * method: init
    def __init__(self, cache: Dict[str, Any] = {}):
        '''
        Initialize the cache context.
        :param cache: An optional initial cache dictionary.
        :type cache: dict
        '''

        # Initialize the cache with the provided dictionary.
        self._cache = cache

    # * method: get
    def get(self, key: str) -> Any:
        '''
        Retrieve an item from the cache.
        :param key: The key of the item to retrieve.
        :type key: str
        :return: The cached item or None if not found.
        :rtype: Any
        '''

        # Return the item from the cache.
        return self._cache.get(key)

    # * method: set
    def set(self, key: str, value: Any):
        '''
        Store an item in the cache.
        :param key: The key to store the value under.
        :type key: str
        :param value: The value to store.
        :type value: Any
        '''

        # Store the value in the cache.
        self._cache[key] = value

    # * method: delete
    def delete(self, key: str):
        '''
        Remove an item from the cache.
        :param key: The key of the item to remove.
        :type key: str
        '''

        # Remove the item from the cache.
        self._cache.pop(key, None)

    # * method: clear
    def clear(self):
        '''
        Clear all items from the cache.
        '''

        # Clear the cache.
        self._cache.clear()



# *** imports

# ** app
from .settings import *


# *** models

# ** model: flagged_dependency
class FlaggedDependency(ValueObject):
    '''
    A flagged container dependency object.
    '''

     # * attribute: module_path
    module_path = StringType(
        required=True,
        metadata=dict(
            description='The module path.'
        )
    )

    # * attribute: class_name
    class_name = StringType(
        required=True,
        metadata=dict(
            description='The class name.'
        )
    )

    # * attribute: flag
    flag = StringType(
        required=True,
        metadata=dict(
            description='The flag for the container dependency.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The container dependency parameters.'
        )
    )


# ** model: container_attribute
class ContainerAttribute(Entity):
    '''
    An attribute that defines container injectior behavior.
    '''

    # * attribute: id
    id = StringType(
        required=True,
        metadata=dict(
            description='The unique identifier for the container attribute.'
        )
    )

    # * attribute: module_path
    module_path = StringType(
        metadata=dict(
            description='The non-flagged module path for the container attribute.'
        )
    )

    # * attribute: class_name
    class_name = StringType(
        metadata=dict(
            description='The non-flagged class name for the container attribute.'
        )
    )

    # * attribute: parameters
    parameters = DictType(
        StringType,
        default={},
        metadata=dict(
            description='The container attribute parameters.'
        )
    )

    # * attribute: dependencies
    dependencies = ListType(
        ModelType(FlaggedDependency),
        default=[],
        metadata=dict(
            description='The container attribute dependencies.'
        )
    )
        
    # * method: get_dependency
    def get_dependency(self, *flags) -> FlaggedDependency:
        '''
        Gets a flagged container dependency by flag.

        :param flags: The flags for the flagged container dependency.
        :type flags: Tuple[str, ...]
        :return: The flagged container dependency that matches the first provided flag.
        :rtype: FlaggedDependency
        '''

        # Return the first dependency that matches any of the provided flags.
        # Input flags are assumed ordinal in priority, so the first match is returned.
        for flag in flags:
            match = next(
                (dependency for dependency in self.dependencies if dependency.flag == flag),
                None
            )
            if match:
                return match
        
        # Return None if no dependency matches the flags.
        return None
        
    # * method: set_dependency
    def set_dependency(self, dependency: FlaggedDependency):
        '''
        Sets a flagged container dependency.

        :param dependency: The flagged container dependency to set.
        :type dependency: ContainerDependency
        '''

        # Replace the value of the dependency if a dependency with the same flag exists.
        for index, dep in enumerate(self.dependencies):
            if dep.flag == dependency.flag:
                self.dependencies[index] = dependency
                return
        
        # Append the dependency otherwise.
        self.dependencies.append(dependency)


# *** imports 

# ** core
from ..domain import *


# *** models

# ** model: error_message0
class ErrorMessage(ValueObject):
    '''
    An error message object.
    '''

    # * attribute: lang
    lang = t.StringType(
        required=True,
        metadata=dict(
            description='The language of the error message text.'
        )
    )

    # * attribute: text
    text = t.StringType(
        required=True,
        metadata=dict(
            description='The error message text.'
        )
    )

    # * method: new
    @staticmethod
    def new(**kwargs) -> 'ErrorMessage':
        '''Initializes a new ErrorMessage object.

        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new ErrorMessage object.
        :rtype: ErrorMessage
        '''

        # Create and return a new ErrorMessage object.
        return super(ErrorMessage, ErrorMessage).new(
            ErrorMessage,
            **kwargs
        )

    # * method: format
    def format(self, *args) -> str:
        '''
        Formats the error message text.

        :param args: The arguments to format the error message text with.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''

        # If there are no arguments, return the error message text.
        if not args:
            return self.text

        # Format the error message text and return it.
        return self.text.format(*args)


# ** model: error
class Error(Entity):
    '''
    An error object.
    '''

    # * attribute: name
    name = t.StringType(
        required=True,
        metadata=dict(
            description='The name of the error.'
        )
    )

    # * attribute: error_code
    error_code = t.StringType(
        metadata=dict(
            description='The unique code of the error.'
        )
    )

    # * attribute: message
    message = t.ListType(
        t.ModelType(ErrorMessage),
        required=True,
        metadata=dict(
            description='The error message translations for the error.'
        )
    )

    # * method: new
    @staticmethod
    def new(name: str, id: str = None, error_code: str = None, **kwargs) -> 'Error':
        '''Initializes a new Error object.

        :param name: The name of the error.
        :type name: str
        :param id: The unique identifier for the error.
        :type id: str
        :param error_code: The error code for the error.
        :type error_code: str
        :param kwargs: Additional keyword arguments.
        :type kwargs: dict
        :return: A new Error object.
        '''

        # Format name as upper case snake case.
        name = name.upper().replace(' ', '_')
        
        # Set Id as the name if not provided.
        if not id:
            id = name

        # Set the error code as the name if not provided.
        if not error_code:
            error_code = name

        # Create and return a new Error object.
        return super(Error, Error).new(
            Error,
            id=id,
            name=name,
            error_code=error_code,
            **kwargs
        )

    # * method: format
    def format(self, lang: str = 'en_US', *args) -> str:
        '''
        Formats the error message text for the specified language.

        :param lang: The language of the error message text.
        :type lang: str
        :param args: The format arguments for the error message text.
        :type args: tuple
        :return: The formatted error message text.
        :rtype: str
        '''

        # Iterate through the error messages.
        for msg in self.message:

            # Skip if the language does not match.
            if msg.lang != lang:
                continue

            # Format the error message text.
            return msg.format(*args)



# *** imports

# ** app
from ..commands import *
from ..contracts.container import *


# *** handlers

# ** handler: container_handler
class ContainerHandler(ContainerService):

    '''
    A container handler is a class that is used to create a container object.
    '''

    # * attribute: container_repo
    container_repo: ContainerRepository

    # * method: __init__
    def __init__(self, container_repo: ContainerRepository):
        '''
        Initialize the container handler.

        :param container_repo: The container repository to use for managing container attributes.
        :type container_repo: ContainerRepository
        '''
        
        # Assign the container repository.
        self.container_repo = container_repo
    
    # * method: list_all
    def list_all(self) -> Tuple[List[ContainerAttribute], Dict[str, str]]:
        '''
        List all container attributes and constants.

        :return: A tuple containing a list of container attributes and a dictionary of constants.
        :rtype: Tuple[List[ContainerAttribute], Dict[str, str]]
        '''

        # Retrieve all container attributes and constants from the repository.
        attributes, constants = self.container_repo.list_all()

        # Return the attributes and constants.
        return attributes, constants
    
    # * method: load_constants
    def load_constants(self, attributes: List[ContainerAttribute], constants: Dict[str, str] = {}, flags: List[str] = []) -> Dict[str, str]:
        '''
        Load constants from the container attributes.

        :param attributes: The list of container attributes.
        :type attributes: List[ContainerAttribute]
        :param constants: The dictionary of constants.
        :type constants: Dict[str, str]
        :return: A dictionary of constants.
        :rtype: Dict[str, str]
        '''

        # Raise an error if there are no attributes provided.
        if not attributes:
            raise_error.execute(
                'CONTAINER_ATTRIBUTES_NOT_FOUND',
                'No container attributes provided to load the container.',
            )

        # If constants are provided, clean the parameters using the parse_parameter command.
        constants = {k: parse_parameter.execute(v) for k, v in constants.items()}

        # Iterate through each attribute.
        for attr in attributes:

            # If flags are provided, check for dependencies with those flags.
            dependency = attr.get_dependency(*flags)

            # Update the constants dictionary with the parsed parameters from the dependency or the attribute itself.
            if dependency:
                constants.update({k: parse_parameter.execute(v) for k, v in dependency.parameters.items()})

            # If no dependency is found, use the attribute's parameters.
            else:
                constants.update({k: parse_parameter.execute(v) for k, v in attr.parameters.items()})

        # Return the updated constants dictionary.
        return constants
    
    # * method: get_dependency_type
    def get_dependency_type(self, attribute: ContainerAttribute, flags: List[str] = []) -> type:
        '''
        Get the type of a container attribute.

        :param attribute: The container attribute.
        :type attribute: ContainerAttribute
        :return: The type of the container attribute.
        :rtype: type
        '''

        # Check the flagged dependencies for the type first.
        for dep in attribute.dependencies:
            if dep.flag in flags:
                return import_dependency.execute(
                    dep.module_path,
                    dep.class_name
                ) 
        
        # Otherwise defer to an available default type.
        if attribute.module_path and attribute.class_name:
            return import_dependency.execute(
                attribute.module_path,
                attribute.class_name
            )
            
        # If no type is found, raise an error.
        raise_error.execute(
            'DEPENDENCY_TYPE_NOT_FOUND',
            f'No dependency type found for attribute {attribute.id} with flags {flags}.',
            attribute.id,
            flags
        )